@@ -24,14 +24,18 @@ from email.parser import Parser
 from pip._vendor.six import StringIO
 
 import pip
+from pip.compat import expanduser
 from pip.download import path_to_url, unpack_url
-from pip.exceptions import InvalidWheelFilename, UnsupportedWheel
+from pip.exceptions import (
+    InstallationError, InvalidWheelFilename, UnsupportedWheel)
 from pip.locations import distutils_scheme, PIP_DELETE_MARKER_FILENAME
 from pip import pep425tags
 from pip.utils import (
-    call_subprocess, ensure_dir, make_path_relative, captured_stdout,
-    rmtree)
+    call_subprocess, ensure_dir, captured_stdout, rmtree, canonicalize_name,
+    read_chunks)
+from pip.utils.ui import open_spinner
 from pip.utils.logging import indent_log
+from pip.utils.setuptools_build import SETUPTOOLS_SHIM
 from pip._vendor.distlib.scripts import ScriptMaker
 from pip._vendor import pkg_resources
 from pip._vendor.six.moves import configparser
@@ -55,7 +59,7 @@ class WheelCache(object):
         :param format_control: A pip.index.FormatControl object to limit
             binaries being read from the cache.
         """
-        self._cache_dir = os.path.expanduser(cache_dir) if cache_dir else None
+        self._cache_dir = expanduser(cache_dir) if cache_dir else None
         self._format_control = format_control
 
     def cached_wheel(self, link, package_name):
@@ -115,7 +119,7 @@ def cached_wheel(cache_dir, link, format_control, package_name):
         return link
     if not package_name:
         return link
-    canonical_name = pkg_resources.safe_name(package_name).lower()
+    canonical_name = canonicalize_name(package_name)
     formats = pip.index.fmt_ctl_formats(format_control, canonical_name)
     if "binary" not in formats:
         return link
@@ -140,7 +144,7 @@ def cached_wheel(cache_dir, link, format_control, package_name):
         return link
     candidates.sort()
     path = os.path.join(root, candidates[0][1])
-    return pip.index.Link(path_to_url(path), trusted=True)
+    return pip.index.Link(path_to_url(path))
 
 
 def rehash(path, algo='sha256', blocksize=1 << 20):
@@ -148,11 +152,9 @@ def rehash(path, algo='sha256', blocksize=1 << 20):
     h = hashlib.new(algo)
     length = 0
     with open(path, 'rb') as f:
-        block = f.read(blocksize)
-        while block:
+        for block in read_chunks(f, size=blocksize):
             length += len(block)
             h.update(block)
-            block = f.read(blocksize)
     digest = 'sha256=' + urlsafe_b64encode(
         h.digest()
     ).decode('latin1').rstrip('=')
@@ -222,6 +224,7 @@ def get_entrypoints(filename):
         data.seek(0)
 
     cp = configparser.RawConfigParser()
+    cp.optionxform = lambda option: option
     cp.readfp(data)
 
     console = {}
@@ -234,12 +237,13 @@ def get_entrypoints(filename):
 
 
 def move_wheel_files(name, req, wheeldir, user=False, home=None, root=None,
-                     pycompile=True, scheme=None, isolated=False):
+                     pycompile=True, scheme=None, isolated=False, prefix=None):
     """Install a wheel"""
 
     if not scheme:
         scheme = distutils_scheme(
-            name, user=user, home=home, root=root, isolated=isolated
+            name, user=user, home=home, root=root, isolated=isolated,
+            prefix=prefix,
         )
 
     if root_is_purelib(name, wheeldir):
@@ -268,7 +272,7 @@ def move_wheel_files(name, req, wheeldir, user=False, home=None, root=None,
         logger.debug(stdout.getvalue())
 
     def normpath(src, p):
-        return make_path_relative(src, p).replace(os.path.sep, '/')
+        return os.path.relpath(src, p).replace(os.path.sep, '/')
 
     def record_installed(srcfile, destfile, modified=False):
         """Map archive RECORD paths to installation RECORD paths."""
@@ -391,6 +395,13 @@ def move_wheel_files(name, req, wheeldir, user=False, home=None, root=None,
     # See https://bitbucket.org/pypa/distlib/issue/34/
     # See https://bitbucket.org/pypa/distlib/issue/33/
     def _get_script_text(entry):
+        if entry.suffix is None:
+            raise InstallationError(
+                "Invalid script entry point: %s for req: %s - A callable "
+                "suffix is required. Cf https://packaging.python.org/en/"
+                "latest/distributing.html#console-scripts for more "
+                "information." % (entry, req)
+            )
         return maker.script_template % {
             "module": entry.prefix,
             "import_name": entry.suffix.split(".")[0],
@@ -486,6 +497,15 @@ if __name__ == '__main__':
             )
         )
 
+    # Record pip as the installer
+    installer = os.path.join(info_dir[0], 'INSTALLER')
+    temp_installer = os.path.join(info_dir[0], 'INSTALLER.pip')
+    with open(temp_installer, 'wb') as installer_file:
+        installer_file.write(b'pip\n')
+    shutil.move(temp_installer, installer)
+    generated.append(installer)
+
+    # Record details of all files installed
     record = os.path.join(info_dir[0], 'RECORD')
     temp_record = os.path.join(info_dir[0], 'RECORD.pip')
     with open_for_csv(record, 'r') as record_in:
@@ -499,7 +519,7 @@ if __name__ == '__main__':
                 writer.writerow(row)
             for f in generated:
                 h, l = rehash(f)
-                writer.writerow((f, h, l))
+                writer.writerow((normpath(f, lib_dir), h, l))
             for f in installed:
                 writer.writerow((installed[f], '', ''))
     shutil.move(temp_record, record)
@@ -654,14 +674,14 @@ class WheelBuilder(object):
         self.build_options = build_options or []
         self.global_options = global_options or []
 
-    def _build_one(self, req, output_dir):
+    def _build_one(self, req, output_dir, python_tag=None):
         """Build one wheel.
 
         :return: The filename of the built wheel, or None if the build failed.
         """
         tempd = tempfile.mkdtemp('pip-wheel-')
         try:
-            if self.__build_one(req, tempd):
+            if self.__build_one(req, tempd, python_tag=python_tag):
                 try:
                     wheel_name = os.listdir(tempd)[0]
                     wheel_path = os.path.join(output_dir, wheel_name)
@@ -669,34 +689,56 @@ class WheelBuilder(object):
                     logger.info('Stored in directory: %s', output_dir)
                     return wheel_path
                 except:
-                    return None
+                    pass
+            # Ignore return, we can't do anything else useful.
+            self._clean_one(req)
             return None
         finally:
             rmtree(tempd)
 
-    def __build_one(self, req, tempd):
-        base_args = [
-            sys.executable, '-c',
-            "import setuptools;__file__=%r;"
-            "exec(compile(open(__file__).read().replace('\\r\\n', '\\n'), "
-            "__file__, 'exec'))" % req.setup_py
+    def _base_setup_args(self, req):
+        return [
+            sys.executable, "-u", '-c',
+            SETUPTOOLS_SHIM % req.setup_py
         ] + list(self.global_options)
 
-        logger.info('Running setup.py bdist_wheel for %s', req.name)
-        logger.debug('Destination directory: %s', tempd)
-        wheel_args = base_args + ['bdist_wheel', '-d', tempd] \
-            + self.build_options
+    def __build_one(self, req, tempd, python_tag=None):
+        base_args = self._base_setup_args(req)
+
+        spin_message = 'Running setup.py bdist_wheel for %s' % (req.name,)
+        with open_spinner(spin_message) as spinner:
+            logger.debug('Destination directory: %s', tempd)
+            wheel_args = base_args + ['bdist_wheel', '-d', tempd] \
+                + self.build_options
+
+            if python_tag is not None:
+                wheel_args += ["--python-tag", python_tag]
+
+            try:
+                call_subprocess(wheel_args, cwd=req.source_dir,
+                                show_stdout=False, spinner=spinner)
+                return True
+            except:
+                spinner.finish("error")
+                logger.error('Failed building wheel for %s', req.name)
+                return False
+
+    def _clean_one(self, req):
+        base_args = self._base_setup_args(req)
+
+        logger.info('Running setup.py clean for %s', req.name)
+        clean_args = base_args + ['clean', '--all']
         try:
-            call_subprocess(wheel_args, cwd=req.source_dir, show_stdout=False)
+            call_subprocess(clean_args, cwd=req.source_dir, show_stdout=False)
             return True
         except:
-            logger.error('Failed building wheel for %s', req.name)
+            logger.error('Failed cleaning build dir for %s', req.name)
             return False
 
     def build(self, autobuilding=False):
         """Build wheels.
 
-        :param unpack: If True, replace the sdist we built from the with the
+        :param unpack: If True, replace the sdist we built from with the
             newly built wheel, in preparation for installation.
         :return: True if all the wheels built correctly.
         """
@@ -733,7 +775,7 @@ class WheelBuilder(object):
                         continue
                     if "binary" not in pip.index.fmt_ctl_formats(
                             self.finder.format_control,
-                            pkg_resources.safe_name(req.name).lower()):
+                            canonicalize_name(req.name)):
                         logger.info(
                             "Skipping bdist_wheel for %s, due to binaries "
                             "being disabled for it.", req.name)
@@ -751,7 +793,9 @@ class WheelBuilder(object):
         with indent_log():
             build_success, build_failure = [], []
             for req in buildset:
+                python_tag = None
                 if autobuilding:
+                    python_tag = pep425tags.implementation_tag
                     output_dir = _cache_for_link(self._cache_root, req.link)
                     try:
                         ensure_dir(output_dir)
@@ -762,7 +806,10 @@ class WheelBuilder(object):
                         continue
                 else:
                     output_dir = self._wheel_dir
-                wheel_file = self._build_one(req, output_dir)
+                wheel_file = self._build_one(
+                    req, output_dir,
+                    python_tag=python_tag,
+                )
                 if wheel_file:
                     build_success.append(req)
                     if autobuilding:
@@ -783,7 +830,7 @@ class WheelBuilder(object):
                             self.requirement_set.build_dir)
                         # Update the link for this.
                         req.link = pip.index.Link(
-                            path_to_url(wheel_file), trusted=True)
+                            path_to_url(wheel_file))
                         assert req.link.is_wheel
                         # extract the wheel into the dir
                         unpack_url(
